import json
import html
from .feature import parse


class BedTable:

    # Always remember the *self* argument
    def __init__(self, bed_file):

        self.features = []

        featureList = parse(bed_file)
        unique_id = 0
        for var in featureList:
            self.features.append((var, unique_id))
            unique_id += 1

    def to_JSON(self):

        headers = ["unique_id", "Chrom", "Start", "End", "Name"]
        rows = []

        for tuple in self.features:
            feature = tuple[0]
            unique_id = tuple[1]
            rows.append([unique_id, feature.chr, feature.start+1, feature.end, html.escape(feature.name)])

        return json.dumps({
            "headers": headers,
            "rows": rows
        })

class BedpeTable:

    # Always remember the *self* argument
    def __init__(self, bedpe_file):

        self.features = []

        featureList = parse(bedpe_file)
        unique_id = 0
        for var in featureList:
            self.features.append((var, unique_id))
            unique_id += 1

    def to_JSON(self):

        headers = ["unique_id", "Chrom_A", "Start_A", "End_A", "Chrom_B", "Start_B", "End_B", "Name"]
        rows = []

        for tuple in self.features:
            feature = tuple[0]
            unique_id = tuple[1]
            rows.append([unique_id, feature.chr, feature.start+1, feature.end, feature.chr2, feature.start2+1, feature.end2, html.escape(feature.name)])

        return json.dumps({
            "headers": headers,
            "rows": rows
        })


class JunctionBedTable:

    # Always remember the *self* argument
    def __init__(self, bed_file, info_columns = None):

        self.features = []
        self.table_columns =  info_columns or None
        featureList = parse(bed_file)
        unique_id = 1
        session_id = 1
        viewport_to_session_id = {}
        for f in featureList:
            #expand name field
            name_tokens = f.name.split(";")
            for token in name_tokens:
                kv = token.split("=")
                key = kv[0]
                value = kv[1]
                setattr(f, key, value)

            # Junction report variants are defined in rows with "viewport" attributes.  Its possible multiple variants
            # share the same viewport, if this is the case they can share a session. This is a space optimization.
            if hasattr(f, 'viewport'):
                self.features.append((f, unique_id))
                unique_id += 1
                viewport = f.viewport
                if viewport in viewport_to_session_id:
                    sid = viewport_to_session_id[viewport]
                else:
                    sid = str(session_id)
                    viewport_to_session_id[viewport] = sid
                    session_id += 1
                f.session_id = sid

    def to_JSON(self):

        json_array = [];
        id_to_session = {}
        for tuple in self.features:

            feature = tuple[0]
            name_tokens = feature.name.split(";")
            unique_id = tuple[1]
            id_to_session[unique_id] = feature.session_id
            obj = {
                "unique_id": unique_id,
                "session_id": feature.session_id,
                "viewport": feature.viewport,
                "feature_locus": feature.chr + ":" + str(feature.start) + "-" + str(feature.end),
                "Chrom": feature.chr,
                "Start": feature.start + 1,
                "End": feature.end
            }

            if self.table_columns == None:
                for token in name_tokens:
                    kv = token.split("=")
                    key = kv[0]
                    if key != 'viewport':
                        value = kv[1]
                        obj[key] = value
            else:
                dict = {}
                for token in name_tokens:
                    kv = token.split("=")
                    dict[kv[0]] = kv[1]
                for key in self.table_columns:
                    if key in dict:
                        obj[key] = dict[key]
                    else:
                        obj[key] = ''


            json_array.append(obj)

        normalized = normalize_json(self.table_columns, json_array)
        normalized["id_to_session"] = id_to_session
        return json.dumps(normalized)


def normalize_json(info_fields, json_array):

    headers = ['unique_id', 'CHROM', 'POSITION', 'REF', 'ALT', 'ID']
    if info_fields is not None:
        for h in info_fields:
            if h == 'ANN':
                headers = headers + [ 'GENE', 'EFFECTS', 'IMPACT', 'TRANSCRIPT', 'GENE_ID', 'PROTEIN ALTERATION', 'DNA ALTERATION']
            else:
                headers.append(h)

    rows = []
    for json in json_array:
        r = []
        for h in headers:
            if h in json:
                r.append(json[h])
            else:
                r.append("")
        rows.append(r)

    return {
        "headers": headers,
        "rows": rows
    }



import unittest
import pathlib
import io
from igv_reports.feature import get_featurereader, FeatureReader, parse_gff, parse_bed


class FeatureFileTest(unittest.TestCase):

    def test_fetch(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        feature_file = get_featurereader(gff)
        content = feature_file.slice()
        features = parse_gff(io.StringIO(content))
        self.assertTrue(len(features) == 2106)

    def test_tabix(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        feature_file = get_featurereader(gff)
        content = feature_file.slice({"chr": "minigenome", "start": 5000, "end": 6000})
        features = parse_gff(io.StringIO(content))
        self.assertEqual(len(features), 33)

        # features should be sorted by start position, and overlap the query interval
        last_start = 0
        for f in features:
            self.assertTrue(f.end >= 5000 and f.start <= 6000)
            self.assertTrue(f.start >= last_start)
            last_start = f.start

    def test_nonindexed_fetch(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        feature_file = FeatureReader(gff)
        content = feature_file.slice()
        features = parse_gff(io.StringIO(content))
        self.assertTrue(len(features) == 2106)

    def test_nonindexed_query(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        feature_file = FeatureReader(gff)
        content = feature_file.slice({"chr": "minigenome", "start": 5000, "end": 6000})
        features = parse_gff(io.StringIO(content))
        self.assertEqual(len(features), 33)

        # features should be sorted by start position, and overlap the query interval
        for f in features:
            self.assertTrue(f.end >= 5000 and f.start <= 6000)

    def test_nongzipped_query(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/variants.bed").resolve())
        feature_file = get_featurereader(gff)
        content = feature_file.slice({"chr": "minigenome", "start": 4000, "end": 7000})
        features = parse_bed(io.StringIO(content))
        self.assertEqual(len(features), 3)

    def test_non_gzipped_query_bedpe(self):

        region = {
            "chr": "chr5",
            "start": 470000,
            "end": 480000
        }
        region2 = {
            "chr": "chr5",
            "start": 180000000,
            "end": 190000000
        }
        bedpe = str((pathlib.Path(__file__).parent / "data/variants.bedpe").resolve())
        feature_file = get_featurereader(bedpe)
        content = feature_file.slice(region, region2=region2)
        features = parse_bed(io.StringIO(content))
        self.assertEqual(len(content.strip().split("\n")), 2)

    def test_multilocus(self):
        path = str((pathlib.Path(__file__).parent / "data/hg38/refGene.txt").resolve())
        feature_reader = get_featurereader(path)
        region = {"chr": "chr5", "start": 470454, "end": 480892}
        region2 = {"chr": "chr7", "start": 55019020, "end": 55019277}
        data = feature_reader.slice(region, region2)
        self.assertEqual(12, count_features(data))

    def test_tabix(self):
        path = str((pathlib.Path(__file__).parent / "data/hg38/refGene.txt.gz").resolve())
        feature_reader = get_featurereader(path)
        region = {"chr": "5", "start": 470454, "end": 480892}
        region2 = {"chr": "7", "start": 55019020, "end": 55019277}
        data = feature_reader.slice(region, region2)
        self.assertEqual(12, count_features(data))

    def test_alias(self):
        path = str((pathlib.Path(__file__).parent / "data/hg38/refGene.txt").resolve())
        feature_reader = get_featurereader(path)
        region = {"chr": "8", "start": 470454, "end": 480892}
        data = feature_reader.slice(region)
        self.assertEqual(2, count_features(data))

    '''
    Test a tabix query for a chromosome that is not present in the data file.
    This should return an array of size zero
    '''
    def test_missingchr(self):
        path = str((pathlib.Path(__file__).parent / "data/hg38/refGene.txt").resolve())
        feature_reader = get_featurereader(path)
        region = {"chr": "999", "start": 470454, "end": 480892}
        data = feature_reader.slice(region)
        self.assertEqual(0, count_features(data))

    '''
    Test a tabix query for a chromosome that is not present in the data file.   
    This should return an array of size zero, and log a warning to the console.
    '''
    def test_tabix_missingchr(self):
        path = str((pathlib.Path(__file__).parent / "data/hg38/refGene.txt.gz").resolve())
        feature_reader = get_featurereader(path)
        region = {"chr": "999", "start": 470454, "end": 480892}
        data = feature_reader.slice(region)
        self.assertEqual(0, count_features(data))

    # Note: the following test is slow
    # def test_remote_tabix(self):
    #     path = "https://s3.amazonaws.com/igv.org.genomes/hg38/annotations/gencode.v28.basic.annotation.sorted.gtf.gz"
    #     feature_reader = FeatureReader(path)
    #     region = {"chr": "5", "start": 470454, "end": 480892}
    #     region2 = {"chr": "7", "start": 55019020, "end": 55019277}
    #     data = feature_reader.slice(region, region2)
    #     print(data)
    #     self.assertEqual(count_features(data), 101)


def count_features(data):
    lines = data.split('\n')
    count = 0
    for line in lines:
        if not line.startswith('#') and len(line) > 0:
            count += 1
    return count



import unittest
import pathlib
from igv_reports import fasta


class FastaTest(unittest.TestCase):

    def test_all(self):

        expected = '>ACACA--STAC2'

        fasta_path = str((pathlib.Path(__file__).parent / 'data/finspector.fa').resolve())
        data = fasta.get_data(fasta_path)

        self.assertTrue(data.startswith(expected))

    def test_region(self):

        region = 'ACACA--STAC2:61-70'
        expected = 'ACAAATATTA'

        data = fasta.get_data('data/finspector.fa', region)

        self.assertTrue(data)
        self.assertEqual(expected, data)

    def test_remote(self):

        data = fasta.get_data("https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg38/hg38.fa",
                              {"chr": "chr1",
                               "start": 100,
                               "end": 200})
        self.assertTrue(data)

    def test_remote_chrom(self):

        data = fasta.get_data("https://s3.amazonaws.com/igv.broadinstitute.org/genomes/seq/hg38/hg38.fa",
                              {"chr": "chr5",
                               "start": 474979,
                               "end": 474998})
        self.assertEqual("ATCAGGCGGCAGAAGGTGCC", data)





def build_aliastable(chrs):

    chralias = {}

    for c in chrs:

        if c.startswith("chr"):
            if c == 'chrM':
                chralias['MT'] = 'chrM'
            else :
                alias = c[3:]
                chralias[alias] = c
        else:
            if c == 'MT':
                chralias['chrM'] = 'MT'
            else:
                alias = f'chr{c}'
                chralias[alias] = c

    return chralias


def get_alias(c):
    if c.startswith("chr"):
        if c == 'chrM':
            return 'chrM'
        else :
            return c[3:]
    else:
        if c == 'MT':
            return 'MT'
        else:
            return f'chr{c}'



from intervaltree import IntervalTree

class FeatureTree:

    def __init__(self, featureList):

        self.featureMap = dict();

        for f in featureList:
            chr = f.chr
            tree = self.featureMap.get(chr)
            if tree == None:
                tree = IntervalTree()
                self.featureMap[chr] = tree
            tree[f.start:f.end] = f

    def query(self, chr, start, end):
        tree = self.featureMap.get(chr)
        if tree == None:
            return set()
        else:
            return tree[start:end]

    def getchrs(self):
        return self.featureMap.keys()



from . import regions
import pysam

def get_data(fasta_file,region=None):

    if None == region:

        with open(fasta_file,"r") as f:
            return(f.read())

    else :

        if isinstance(region,str):
            region = regions.parse_region(region)

        chr = region["chr"]
        start = region["start"] - 1
        end = region["end"]

        fasta = pysam.FastaFile(fasta_file)

        slice_seq = fasta.fetch(chr, start, end)

        fasta.close()

        return slice_seq

class FastaReader:

    def __init__(self, path):

        self.fasta = pysam.FastaFile(path)

    def slice(self, region = None):

        if None == region:
            with open(self.fasta,"r") as f:
                return(f.read())

        else :

            if isinstance(region,str):
                region = regions.parse_region(region)

            chr = region["chr"]
            start = region["start"] - 1
            end = region["end"]

            try:
                seq = self.fasta.fetch(chr, start, end)


            except KeyError:
                chr = "chr" + chr
                seq = self.fasta.fetch(chr, start, end)

            return seq





import unittest
import pathlib
import types

from igv_reports.bam import BamReader

class BAMTest(unittest.TestCase):


    def test_bam(self):

        region = {
            "chr": "minigenome",
            "start": 4000,
            "end": 10000
        }

        bam_file_path = str((pathlib.Path(__file__).parent / "data/minigenome/alignments.bam").resolve())
        bamreader = BamReader("bam", bam_file_path)
        data = bamreader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 923)

    def test_bam_noregion(self):

        bam_file_path = str((pathlib.Path(__file__).parent / "data/minigenome/alignments.bam").resolve())
        bamreader = BamReader("bam", bam_file_path)
        data = bamreader.slice(region=None, sam=True)
        self.assertEqual(count_alignments(data), 10212)


    def test_cram(self):
        region = {
            "chr": "minigenome",
            "start": 4000,
            "end": 10000
        }

        bam_file_path = str((pathlib.Path(__file__).parent / "data/minigenome/alignments.cram").resolve())
        ref_file_path = str((pathlib.Path(__file__).parent / "data/minigenome/minigenome.fa").resolve())

        args =  types.SimpleNamespace()
        args.fasta = ref_file_path
        reader = BamReader("cram", bam_file_path, args);
        data = reader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 923)


    def test_multiple_bam_regions_diff_chrom(self):
        region = {
            "chr": "chr5",
            "start": 474989,
            "end": 474989
        }
        region2 = {
            "chr": "chr8",
            "start": 143923759,
            "end": 143923759
        }

        bam_file_path = str((pathlib.Path(__file__).parent / "data/variants/recalibrated.bam").resolve())
        reader = BamReader("bam", bam_file_path);
        data = reader.slice(region, region2=region2, sam=True)
        self.assertEqual(count_alignments(data), 81)


    def test_multiple_bam_regions_same_chrom(self):
        region = {
            "chr": "chr5",
            "start": 474989,
            "end": 474989
        }
        region2 = {
            "chr": "chr5",
            "start": 181224474,
            "end": 181224474
        }

        bam_file_path = str((pathlib.Path(__file__).parent / "data/variants/recalibrated.bam").resolve())
        reader = BamReader("bam", bam_file_path);
        data = reader.slice(region, region2=region2, sam=True)
        self.assertEqual(count_alignments(data), 58)

    def test_chralias(self):

        region = {
            "chr": "5",
            "start": 474989,
            "end": 474989
        }

        region2 = {
            "chr": "5",
            "start": 181224474,
            "end": 181224474
        }

        bam_file_path = str((pathlib.Path(__file__).parent / "data/variants/recalibrated.bam").resolve())
        reader = BamReader("bam", bam_file_path);
        data = reader.slice(region, region2, sam=True)
        self.assertEqual(count_alignments(data), 58)

    def test_exclude_flags(self):

        args =  types.SimpleNamespace()
        args.exclude_flags = 1536
        bam_file_path = str((pathlib.Path(__file__).parent / "data/dups/dups.bam").resolve())
        bamreader = BamReader("bam", bam_file_path, args)
        region = {
            "chr": "1",
            "start": 658371,
            "end": 658460
        }
        data = bamreader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 73)

        args.exclude_flags = 1024
        bamreader = BamReader("bam", bam_file_path, args)
        data = bamreader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 73)

        args.exclude_flags = 0
        bamreader = BamReader("bam", bam_file_path, args)
        data = bamreader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 121)

        args.exclude_flags = 512
        bamreader = BamReader("bam", bam_file_path, args)
        data = bamreader.slice(region, sam=True)
        self.assertEqual(count_alignments(data), 121)



def count_alignments(data):

    lines = data.split('\n')
    count = 0
    for line in lines:
        if len(line) > 0 and not line.startswith("@"):
            count += 1
    return count



import unittest
import pathlib
import types

from igv_reports.wig import WigReader

class WIGTest(unittest.TestCase):

    def test_wig(self):

        region = {
            "chr": "chr22",
            "start": 42126499,
            "end": 42130810
        }

        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/variable_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region)
        self.assertEqual(count_lines(data), 177)

    def test_wig_noregion(self):
        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/variable_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region=None)
        self.assertEqual(count_lines(data), 1284)

    def test_multiple_wig_regions_diff_chrom(self):
        region = {
            "chr": "chr7",
            "start": 99756967,
            "end": 99784184
        }
        region2 = {
            "chr": "chr22",
            "start": 42126499,
            "end": 42130810
        }

        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/variable_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region, region2)
        self.assertEqual(count_lines(data), 1268)


    def test_multiple_wig_regions_same_chrom(self):
        region = {
            "chr": "chr22",
            "start": 42126499,
            "end": 42128000
        }
        region2 = {
            "chr": "chr22",
            "start": 42128000,
            "end": 42130810
        }

        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/variable_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region, region2)
        self.assertEqual(count_lines(data), 64)
    
    def test_wig_mixed_step_same_chrom(self):
        region = {
            "chr": "chr19",
            "start": 49306691,
            "end": 49309000
        }

        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/mixed_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region)
        self.assertEqual(count_lines(data), 10)


    def test_chralias(self):
        region = {
            "chr": "7",
            "start": 99756967,
            "end": 99784184
        }
        region2 = {
            "chr": "22",
            "start": 42126499,
            "end": 42130810
        }

        wig_file_path = str((pathlib.Path(__file__).parent / "data/wig/variable_step.wig").resolve())
        wigreader = WigReader(wig_file_path)
        data = wigreader.slice(region, region2)
        self.assertEqual(count_lines(data), 1268)

def count_lines(data):
    return len([i for i in data.split('\n')])